' This program is provided as Freeware for personal, educational
' and non-commercial use under a GNU General Public License.
'
' Written by Mike Perks (basicx@austin.rr.com), 2005
'
' Devantech SRF10/08 interface module
' (see http://www.robot-electronics.co.uk/shop/Ultrasonic_Rangers1999.htm)
'
' Written for ZBasic Application Note AN-208
'
Option Explicit
Option Strict On
Option TaskStackMargin 5

' *************************************************************
' Define Public Constant Values for SRF10 and SRF08
'
' Note that some constants are different for the SRF10 and SRF08
' *************************************************************
Public Const SRF_UNITS_INCHES as Byte 		= 80
Public Const SRF_UNITS_CM as Byte 			= 81
Public Const SRF_UNITS_MICROSECONDS as Byte	= 82

Public Const SRF10_GAIN_40 as Byte	= 1
Public Const SRF10_GAIN_50 as Byte	= 2
Public Const SRF10_GAIN_60 as Byte	= 3
Public Const SRF10_GAIN_70 as Byte	= 4
Public Const SRF10_GAIN_80 as Byte	= 5
Public Const SRF10_GAIN_100 as Byte	= 6
Public Const SRF10_GAIN_120 as Byte	= 7
Public Const SRF10_GAIN_140 as Byte	= 8
Public Const SRF10_GAIN_200 as Byte	= 9
Public Const SRF10_GAIN_250 as Byte	= 10
Public Const SRF10_GAIN_300 as Byte	= 11
Public Const SRF10_GAIN_350 as Byte	= 12
Public Const SRF10_GAIN_400 as Byte	= 13
Public Const SRF10_GAIN_500 as Byte	= 14
Public Const SRF10_GAIN_600 as Byte	= 15
Public Const SRF10_GAIN_700 as Byte	= 16
Public Const SRF10_DEFAULT_GAIN as Byte = SRF10_GAIN_140

Public Const SRF08_GAIN_94   as Byte = 0  
Public Const SRF08_GAIN_97   as Byte = 1  
Public Const SRF08_GAIN_100  as Byte = 2  
Public Const SRF08_GAIN_103  as Byte = 3  
Public Const SRF08_GAIN_107  as Byte = 4  
Public Const SRF08_GAIN_110  as Byte = 5  
Public Const SRF08_GAIN_114  as Byte = 6  
Public Const SRF08_GAIN_118  as Byte = 7  
Public Const SRF08_GAIN_123  as Byte = 8  
Public Const SRF08_GAIN_128  as Byte = 9  
Public Const SRF08_GAIN_133  as Byte = 10 
Public Const SRF08_GAIN_139  as Byte = 11 
Public Const SRF08_GAIN_145  as Byte = 12 
Public Const SRF08_GAIN_152  as Byte = 13 
Public Const SRF08_GAIN_159  as Byte = 14 
Public Const SRF08_GAIN_168  as Byte = 15 
Public Const SRF08_GAIN_177  as Byte = 16 
Public Const SRF08_GAIN_187  as Byte = 17 
Public Const SRF08_GAIN_199  as Byte = 18 
Public Const SRF08_GAIN_212  as Byte = 19 
Public Const SRF08_GAIN_227  as Byte = 20 
Public Const SRF08_GAIN_245  as Byte = 21 
Public Const SRF08_GAIN_265  as Byte = 22 
Public Const SRF08_GAIN_288  as Byte = 23 
Public Const SRF08_GAIN_317  as Byte = 24 
Public Const SRF08_GAIN_352  as Byte = 25 
Public Const SRF08_GAIN_395  as Byte = 26 
Public Const SRF08_GAIN_450  as Byte = 27 
Public Const SRF08_GAIN_524  as Byte = 28 
Public Const SRF08_GAIN_626  as Byte = 29 
Public Const SRF08_GAIN_777  as Byte = 30 
Public Const SRF08_GAIN_1025 as Byte = 31 
Public Const SRF08_DEFAULT_GAIN as Byte = SRF08_GAIN_139

' *************************************************************
' Sub InitSRF
' 
' Initialize the SRF10/SRF08 by configuring it and creating a new task
' Parm1 is the I2C address of the SRF
' Parm2 is the required maximum range in inches or cmd
' Parm3 is the units for parm2 (inches or cm)
' Parm4 is the gain setting for the SRF - see SRFxx_GAIN_xxx constants
'**************************************************************
Public Sub InitSRF(ByVal addr as Byte, ByVal reqRange as UnsignedInteger, _
					ByVal units as Byte, ByVal gain as Byte)
	' save data for later use
	address = addr

	' open the I2C channel and configure the gain
	Call OpenI2C(channel, sdaPin, sclPin)
	Call setSRFGain(addr, gain)
	
	' initialize the maximum range and start the range task
	maximumRange = setSRFMaxRange(addr, reqRange, units)
	range = 0
	CallTask "doRanging", rangeStack  
	Call Sleep(0.1)
End Sub	

' *************************************************************
' Sub TermSRF
' 
' Terminate using the SF10/SRF08 by exiting the range task and
' closing the I2C channel
'**************************************************************
Public Sub TermSRF()
	Call ExitTask(rangeStack)
	Call CloseI2C(channel)
End Sub

' *************************************************************
' Function GetSRFRange
' 
' Returns the last range reading in a unsigned integer
'**************************************************************
Public Function GetSRFRange() as UnsignedInteger
	If range = 0 Then
		GetSRFRange = maximumRange
	Else	
		GetSRFRange = range
	End If	
End Function	

' *************************************************************
' Function GetSRFVersion
' 
' Returns the software version of the SRF in a byte
' Parm1 is the I2C address of the SRF
'**************************************************************
Public Function GetSRFVersion(ByVal addr as Byte) As Byte
	Dim rc as Integer
	Dim reg as Byte
	reg = REG_REV
	rc = I2CCmd(channel, addr, CByte(SizeOf(reg)), reg, 1, GetSRFVersion)
	If rc < 1 Then
		Debug.Print "GetSRFVersion i2cmd returned ";CStr(rc)
		GetSRFVersion = 0
	End If	
End Function

' *************************************************************
' Function SetSRFAddress
' 
' Sets a new address for the SRF and returns the current address
' Parm1 is the I2C address of the SRF
' Parm2 is the new I2C address
'**************************************************************
Public Function SetSRFAddress(ByVal addr as Byte, ByVal newAddr as Byte) as Byte
	SetSRFAddress = addr
	If sendCmd(addr, REG_CMD, &HA0) < 0 Then
		Debug.Print "SetSRFaddr i2cmd A0 returned error"
		Exit Function
	End If	
	If sendCmd(addr, REG_CMD, &HAA) < 0 Then
		Debug.Print "SetSRFaddr i2cmd AA returned error"
		Exit Function
	End If	
	If sendCmd(addr, REG_CMD, &HA5) < 0 Then
		Debug.Print "SetSRFaddr i2cmd A5 returned error"
		Exit Function
	End If	
	If sendCmd(addr, REG_CMD, newAddr) < 0 Then
		Debug.Print "SetSRFaddr i2cmd set address returned error"
		Exit Function
	End If	
	SetSRFAddress = newAddr
End Function


' *************************************************************
' Define Private Constant Values for SRF10 and SRF08
' *************************************************************

' Define SRF10 Registers
Private Const REG_CMD  As Byte = &H00		' Command Register (Write)
Private Const REG_REV  As Byte = &H00		' Revision Register (Read)
Private Const REG_GAIN As Byte = &H01  		' Gain Register (Write)
Private Const REG_RANGE As Byte = &H02  	' Range Setting Byte (Write)
Private Const REG_RANGE_HIGH As Byte = &H02	' Range High Byte (Read)
Private Const REG_RANGE_LOW  As Byte = &H03	' Range Low Byte (Read)

' Define other internal constants
Private Const MAX_RANGE as Byte = 255
Private Const MAX_DISTANCE_CM as UnsignedInteger = 1100
Private Const MAX_DISTANCE_INCHES as UnsignedInteger = 433  ' 1100/2.54
Private Const PER_UNIT_RANGE as Single		= 43.0	' 43 mm 
Private Const RANGING_WAIT_TIME as Single 	= 0.065	' Time to wait for ranging to complete

' I2C channel constants
Private Const channel as Byte = 0
Private Const sdaPin as Byte = C.1
Private Const sclPin as Byte = C.0

' private data for SRF
Private address as Byte
Private range as UnsignedInteger
Private maximumRange as UnsignedInteger
Private rangeMode as Byte

' task stack for range task
Private Const STACK_SIZE as Byte = 36
Private rangeStack(1 to STACK_SIZE) as Byte

' *************************************************************
' Private Sub setSRFGain
' 
' Sets the SRF gain to the given value for the SRF at the given I2C address
' The gain should be one of the corresponding SRFxx_GAIN_xxx constants
' This routine should not be called while the SRF is ranging
' Parm1 is the I2C address of the SRF
' Parm2 is the gain setting for the SRF - see SRFxx_GAIN_xxx constants
'**************************************************************
Private Sub setSRFGain(ByVal addr as Byte, ByVal gain As Byte)
	If sendCmd(addr, REG_GAIN, gain) < 0 Then
		Debug.Print "SetSRFGain i2cmd returned error"
	End If	
End Sub

' *************************************************************
' Private Sub setSRFMaxRange
' 
' Sets the maximum range for the SRF
' This routine should not be called while the SRF is ranging
' Parm1 is the I2C address of the SRF
' Parm2 is the required maximum range in inches or cmd
' Parm3 is the units for parm2 and the rangeMode
'**************************************************************
Private Function setSRFMaxRange(ByVal addr as Byte, ByVal reqRange as UnsignedInteger, ByVal units as Byte) as UnsignedInteger
	rangeMode = units
	If sendCmd(addr, REG_RANGE, calculateRange(reqRange, units)) < 0 Then
		Debug.Print "SetSRFRange i2cmd returned error"
		maximumRange = 0
	Else
		maximumRange = reqRange
	End If	
	setSRFMaxRange = maximumRange
End Function

' *************************************************************
' Private Sub doRanging
' 
' This routine loops forever getting the range from the SRF
' and setting the private range variable
' This routine is normally ran as a separate task
'**************************************************************
Private Sub doRanging()
	Dim cmd(1 to 2) as Byte
	Dim reg as Byte
	Dim rc as Integer
	
	Do	
		' start ranging
		cmd(1) = REG_CMD
		cmd(2) = rangeMode
		rc = I2CCmd(channel, address, CByte(SizeOf(cmd)), cmd, 0, 0)
		If rc < 0 Then
			Debug.Print "GetRange start i2cmd returned ";CStr(rc)
		End if

		' sleep while waiting for the ranging to complete
		Call Sleep(RANGING_WAIT_TIME)
		
		reg = REG_RANGE_HIGH
		rc = I2CCmd(channel, address, 1, reg, 2, cmd)
		If rc = 2 Then
           ' the MS byte is returned in the first byte, LS byte in the second
           range = MakeWord(cmd(2), cmd(1))
		Else
           Debug.Print "GetRange i2cmd returned ";CStr(rc)
		End If
	Loop
End Sub

' *************************************************************
' Private Function sendCmd
' 
' This routine sends a single byte command to a SRF register
' Parm1 is the I2C address of the SRF
' Parm2 is the SRF register number
' Parm3 is the value to write into the register
'**************************************************************
Private Function sendCmd(ByVal addr as Byte, ByVal reg as Byte, ByVal command as Byte) as Integer
	Dim cmd(1 to 2) as Byte	
	cmd(1) = reg
	cmd(2) = command
	SendCmd = I2CCmd(channel, addr, CByte(SizeOf(cmd)), cmd, 0, 0)
End Function

' *************************************************************
' Private Function calculateRange
' 
' This routine converts a range distance to a number for the range register
' Parm1 is the required range and is also updated with the actual range
' Parm2 is the units defined as inches or centimeters
'**************************************************************
Private Function calculateRange(ByRef maxRange as UnsignedInteger, ByVal units as Byte) as Byte
	If units = SRF_UNITS_CM Then
		If maxRange > MAX_DISTANCE_CM Then
			calculateRange = MAX_RANGE
		Else	
			calculateRange = CByte(CSng(maxRange) * 10.0 / PER_UNIT_RANGE)
		End If
		maxRange = CUInt((CSng(calculateRange) + 1.0) * (PER_UNIT_RANGE / 10.0))
	Else
		' default to inches
		If maxRange > MAX_DISTANCE_INCHES Then
			calculateRange = MAX_RANGE
		Else	
			calculateRange = CByte(CSng(maxRange) * 25.4 / PER_UNIT_RANGE)
		End If	
		maxRange = CUint((CSng(calculateRange) + 1.0) * (PER_UNIT_RANGE / 25.4))
	End If	
End Function